<?php

namespace Drupal\denormalizer\Entity;

use Drupal\Core\Config\Entity\ConfigEntityInterface;
use Drupal\Core\Entity\EntityInterface;

/**
 * Provides an interface for defining Denormalizer table entities.
 */
interface DenormalizerTableInterface extends ConfigEntityInterface {

  /**
   * Gets the Denormalizer table source ID.
   *
   * @return string
   *   The Denormalizer table source ID.
   */
  public function getSourceId(): string;

  /**
   * Sets the Denormalizer table source ID.
   *
   * @param string $source_id
   *   The Denormalizer table source ID.
   *
   * @return $this
   */
  public function setSourceId($source_id);

  /**
   * Gets the Denormalizer table source configuration.
   *
   * @return array
   *   The Denormalizer table source configuration.
   */
  public function getSourceConfiguration(): array;

  /**
   * Sets the Denormalizer table source configuration.
   *
   * @param array $configuration
   *   The Denormalizer table source configuration.
   *
   * @return $this
   */
  public function setSourceConfiguration(array $configuration);

  /**
   * Gets the Denormalizer table source entity type.
   *
   * @return string
   *   The Denormalizer table source entity type.
   */
  public function getDbTableName(): string;

  /**
   * Gets the Denormalizer base table name.
   *
   * @return string
   *   The Denormalizer base table name.
   */
  public function getBaseTable(): string;

  /**
   * Creates the Denormalizer database table.
   *
   * @param bool $init
   *   The flag to initialize the table.
   *
   * @return bool
   *   The Denormalizer table.
   */
  public function createDatabaseTable(bool $init = FALSE): bool;

  /**
   * Inserts data from an entity into the database table.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity to insert into the database table.
   *
   * @return int|string|null
   *   The inserted row ID.
   */
  public function insertDataFromEntityIntoDbTable(EntityInterface $entity): int|string|null;

  /**
   * Updates data from an entity into the database table.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity to update into the database table.
   */
  public function updateDataFromEntityIntoDbTable(EntityInterface $entity);

  /**
   * Deletes data from an entity into the database table.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity to delete from the database table.
   */
  public function deleteDataFromEntityIntoDbTable(EntityInterface $entity);

  /**
   * Gets the schema for the database table.
   */
  public function getDbTableFieldsSchema() : array;

  /**
   * Checks if the source is a webform submission.
   */
  public function isWebformSubmission(): bool;

  /**
   * Gets the entity selected fields.
   *
   * @param bool $base_field_only
   *   The flag to get only the base fields.
   *
   * @return array
   *   The entity selected fields.
   */
  public function getEntitySelectedFields(bool $base_field_only = FALSE): array;

  /**
   * Gets the webform submission selected elements.
   *
   * @return array
   *   The webform submission selected elements.
   */
  public function getWebformSubmissionSelectedElements(): array;

  /**
   * Counts the number of rows.
   *
   * @return int
   *   The number of rows in the table.
   */
  public function countRows(): int;

}
