<?php

namespace Drupal\denormalizer\Form;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Database\Database;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Denormalizer Settings Form.
 */
class DenormalizerSettingsForm extends ConfigFormBase {

  /**
   * The entity bundle service.
   *
   * @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface
   */
  protected $bundleInfo;

  /**
   * DenormalizerSettingsForm constructor.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   Config factory.
   * @param \Drupal\Core\Entity\EntityTypeBundleInfoInterface $bundleInfo
   *   The bundle info.
   */
  public function __construct(ConfigFactoryInterface $configFactory, EntityTypeBundleInfoInterface $bundleInfo) {
    parent::__construct($configFactory);
    $this->bundleInfo = $bundleInfo;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('entity_type.bundle.info')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'denormalizer_settings_form';
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return ['denormalizer.settings'];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $config = $this->config('denormalizer.settings');

    $form['sql_mode'] = [
      '#title' => t('SQL mode'),
      '#type' => 'radios',
      '#options' => [
        'views' => t('Create views'),
        'tables' => t('Create tables'),
      ],
      '#description' => t('Use views or tables for generating normalized data.'),
      '#default_value' => $config->get('sql_mode'),
    ];

    $form['db'] = [
      '#title' => t('DB'),
      '#type' => 'radios',
      '#options' => [
        'local' => t('Use local database'),
        'external' => t('Create external database'),
      ],
      '#description' => t('Which database to use'),
      '#default_value' => $config->get('db'),
    ];

    $form['view_prefix'] = [
      '#title' => t('View prefix'),
      '#type' => 'textfield',
      '#default_value' => $config->get('view_prefix'),
    ];

    $form['db_prefix'] = [
      '#title' => t('Database prefix'),
      '#type' => 'textfield',
      '#default_value' => $config->get('db_prefix'),
    ];

    $form['cron_enabled'] = [
      '#title' => t('Run on cron?'),
      '#type' => 'checkbox',
      '#default_value' => $config->get('cron_enabled'),
      '#description' => t('The denormalized views or tables will be updated incrementally on cron.'),
    ];

    $form['run_every'] = [
      '#title' => t('Run every'),
      '#type' => 'textfield',
      '#default_value' => $config->get('run_every'),
      '#description' => t('Time between runs.'),
    ];

    $form['reload_every'] = [
      '#title' => t('Reload every'),
      '#type' => 'textfield',
      '#default_value' => $config->get('reload_every'),
      '#description' => t('Time between reloads.'),
    ];

    if (\Drupal::moduleHandler()->moduleExists('duration_field')) {
      /** @var \Drupal\duration_field\Service\DurationService $duration */
      $duration = \Drupal::service('duration_field.service');

      $form['run_every']['#type'] = 'duration';
      $form['run_every']['#default_value'] = $duration->getDurationStringFromDateInterval($this->secondsToInterval($config->get('run_every')));
      $form['run_every']['#granularity'] = 'd:h:i';

      $form['reload_every']['#type'] = 'duration';
      $form['reload_every']['#granularity'] = 'd:h:i';
      $form['reload_every']['#default_value'] = $duration->getDurationStringFromDateInterval($this->secondsToInterval($config->get('reload_every')));
    }

    return parent::buildForm($form, $form_state);
  }

  /**
   * Convert seconds to full ISO8601 string.
   *
   * Yes, we can use DateInterval but Duration expects a full string
   * (not just PT100S).
   *
   * @param int|string $seconds
   *   The seconds.
   *
   * @return \DateInterval
   *   The date interval.
   */
  public function secondsToInterval(int|string $seconds): \DateInterval {
    $seconds = (int) $seconds;
    $date1 = new \DateTime('1904-01-01');
    $date2 = new \DateTime('1904-01-01');
    $date2->add(new \DateInterval("PT{$seconds}S"));
    return $date2->diff($date1);
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    parent::validateForm($form, $form_state);

    // Get the prefix from the default database.
    $prefix = '';
    if ($form_state->getValue('db') == 'local') {
      $connection = Database::getConnection();
      $prefix = $connection->getPrefix();
    }

    $no_prefix = $form_state->getValue('db') == 'local' && !$form_state->getValue('view_prefix');
    $same_prefix = $form_state->getValue('db') == 'local' && $form_state->getValue('view_prefix') == $prefix;
    if ($no_prefix || $same_prefix) {
      $form_state->setErrorByName('view_prefix', $this->t('Using the local database requires a view prefix. Otherwise your tables will be overwritten!'));
    }

    if ($form_state->getValue('db') == 'external' && !$form_state->getValue('db_prefix')) {
      $form_state->setErrorByName('db_prefix', $this->t('Database prefix is required if using an external database.'));
    }

    if (\Drupal::moduleHandler()->moduleExists('duration_field')) {
      /** @var \Drupal\duration_field\Service\DurationService $duration */
      $duration = \Drupal::service('duration_field.service');

      $form_state->setValue('run_every', $duration->getSecondsFromDateInterval($form_state->getValue('run_every')));
      $form_state->setValue('reload_every', $duration->getSecondsFromDateInterval($form_state->getValue('reload_every')));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $this->config('denormalizer.settings')
      ->setData($form_state->getValues())
      ->save();
    parent::submitForm($form, $form_state);
  }

}
