<?php

namespace Drupal\denormalizer\Form;

use Drupal\Component\Utility\Html;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\denormalizer\Entity\DenormalizerTableInterface;
use Drupal\webform\WebformInterface;

/**
 * Denormalizer table form.
 */
class DenormalizerTableForm extends EntityForm {

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $form = parent::form($form, $form_state);
    /** @var \Drupal\denormalizer\Entity\DenormalizerTableInterface $table */
    $table = $this->entity;
    $sources = [
      'entity' => $this->t('Entity'),
      'non_entity' => $this->t('Non entity'),
    ];

    // Use the "entity" as first table source for the default value.
    if (!$table->getSourceId()) {
      $source_ids = array_keys($sources);
      $source = reset($source_ids);
      $table->setSourceId($source);
    }
    // The form state will have a plugin value if #ajax was used.
    $source = $form_state->getValue('source', $table->getSourceId());
    // Pass the plugin configuration only if the plugin hasn't been changed
    // via #ajax.
    $source_configuration = $table->getSourceId() == $source ? $table->getSourceConfiguration() : [];

    $wrapper_id = Html::getUniqueId('denormalizer-table-form');
    $form['#prefix'] = '<div id="' . $wrapper_id . '">';
    $form['#suffix'] = '</div>';
    $form['#tree'] = TRUE;

    $form['label'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Name'),
      '#maxlength' => 255,
      '#default_value' => $table->label(),
      '#required' => TRUE,
    ];
    $form['id'] = [
      '#type' => 'machine_name',
      '#default_value' => $table->id(),
      '#machine_name' => [
        'exists' => '\Drupal\denormalizer\Entity\DenormalizerTable::load',
      ],
      '#disabled' => !$table->isNew(),
    ];
    $form['source'] = [
      '#type' => 'select',
      '#title' => $this->t('Source'),
      '#options' => $sources,
      '#default_value' => $source,
      '#required' => TRUE,
      '#disabled' => !$table->isNew(),
      '#ajax' => [
        'callback' => '::ajaxRefresh',
        'wrapper' => $wrapper_id,
      ],
    ];

    $form['configuration'] = [
      '#type' => 'container',
      '#tree' => TRUE,
      '#denormalizer_table_source_id' => $source,
      '#denormalizer_table_configuration' => $source_configuration,
    ];
    $form['configuration'] = $this->buildSourceConfigurationForm($form['configuration'], $form_state);

    $form['status'] = [
      '#type' => 'radios',
      '#title' => $this->t('Status'),
      '#options' => [
        0 => $this->t('Disabled'),
        1 => $this->t('Enabled'),
      ],
      '#default_value' => (int) $table->status(),
    ];

    return $form;
  }

  /**
   * Ajax callback to refresh the main form.
   */
  public static function ajaxRefresh(array $form, FormStateInterface $form_state) {
    $triggering_element = $form_state->getTriggeringElement();
    if (!empty($triggering_element['#configuration_sub_form'])) {
      return $form['configuration'];
    }
    elseif (!empty($triggering_element['#configuration_bundle_refresh'])) {
      return $form['configuration']['fields']['bundle'];
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    parent::submitForm($form, $form_state);
    /** @var \Drupal\denormalizer\Entity\DenormalizerTableInterface $table */
    $table = $this->entity;
    $table->setSourceConfiguration($form_state->getValue(['configuration']));
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    $this->entity->save();

    if ($this->entity->getSourceId() !== 'entity') {
      $this->messenger()
        ->addWarning($this->t('Only @entity config entity was created! None entity source are not currently supported!'));
      $form_state->setRedirect('entity.denormalizer_table.collection');
      return;
    }

    $configuration = $this->entity->getSourceConfiguration();
    $entity_type_id = $configuration['entity_type'];
    $entity_storage = $this->entityTypeManager->getStorage($entity_type_id);

    $class = get_class($this->entity);
    $operations[] = [
      [$class, 'initBatch'],
      [$this->entity, $entity_type_id],
    ];

    $entity_type_definition = $this->entityTypeManager->getDefinition($entity_type_id);
    $entity_ids = $this->getEntityIdsToPopulate($this->entity, $entity_storage, $entity_type_definition);
    $entity_ids_chunks = array_chunk($entity_ids, 1000);
    foreach ($entity_ids_chunks as $entity_ids_chunk) {
      $operations[] = [
        [$class, 'populateTableBatch'],
        [$entity_ids_chunk],
      ];
    }

    $batch = [
      'title' => $this->t('Initializing and populating the denormalized table'),
      'operations' => $operations,
      'finished' => [$class, 'finishBatch'],
    ];
    batch_set($batch);
  }

  /**
   * Build the source confirguration form.
   *
   * @param array $sub_form
   *   The sub form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   *
   * @return array
   *   The source configuration form.
   */
  private function buildSourceConfigurationForm($sub_form, FormStateInterface $form_state): array {
    $source_id = $sub_form['#denormalizer_table_source_id'];
    $configuration = $sub_form['#denormalizer_table_configuration'];
    $wrapper_id = Html::getUniqueId('denormalizer-table-form--configuration');
    $sub_form['#prefix'] = '<div id="' . $wrapper_id . '">';
    $sub_form['#suffix'] = '</div>';

    if ($source_id === 'entity') {
      /** @var \Drupal\Core\Entity\EntityTypeRepositoryInterface $entity_type_repository */
      $entity_type_repository = \Drupal::service('entity_type.repository');
      $entity_type = $form_state->getValue([
        'configuration',
        'entity_type',
      ], '');
      if (!$entity_type) {
        $entity_type = $configuration['entity_type'] ?? '';
      }
      $entities = $entity_type_repository->getEntityTypeLabels(TRUE)['Content'];
      $sub_form['entity_type'] = [
        '#type' => 'select',
        '#title' => $this->t('Entity type'),
        '#options' => ['' => '- None -'] + $entities,
        '#default_value' => $entity_type,
        '#required' => TRUE,
        '#configuration_sub_form' => TRUE,
        '#ajax' => [
          'callback' => '::ajaxRefresh',
          'wrapper' => $wrapper_id,
        ],
      ];

      if ($entity_type) {
        $bundle = $form_state->getValue(['configuration', 'bundle'], '');
        $bundles = $this->getBundleOptions($entity_type);
        if ($bundle && !isset($bundles[$bundle])) {
          $bundle = '';
        }
        elseif (!$bundle) {
          $bundle = $configuration['bundle'] ?? '';
        }
        $wrapper_id_bundle_fields = $wrapper_id . '--entity-bundle-fields';
        $sub_form['bundle'] = [
          '#type' => 'select',
          '#default_value' => $bundle,
          '#title' => $this->t('Bundle'),
          '#options' => ['' => '- None -'] + $bundles,
          '#required' => TRUE,
          '#configuration_bundle_refresh' => TRUE,
          '#ajax' => [
            'callback' => '::ajaxRefresh',
            'wrapper' => $wrapper_id_bundle_fields,
          ],
        ];
        $sub_form['changed_key'] = [
          '#type' => 'select',
          '#title' => $this->t('Changed key'),
          '#default_value' => $configuration['changed_key'] ?? '',
          '#options' => ['' => '- None -'] + $this->getEntityBaseFieldsOptions($entity_type, TRUE),
        ];

        $sub_form['fields'] = [
          '#type' => 'details',
          '#title' => $this->t('Fields'),
        ];
        $sub_form['fields']['base'] = [
          '#type' => 'checkboxes',
          '#title' => $this->t('Base fields'),
          '#field_prefix' => $this->t('<div class="description fieldset__description">Please select the base fields to include into the denormalized table. Leave empty to include all of them.</div>'),
          '#default_value' => $configuration['fields']['base'] ?? [],
          '#options' => $this->getEntityBaseFieldsOptions($entity_type),
        ];
        $this->processEntityKeysOptions($sub_form['fields']['base'], $entity_type);

        $sub_form['fields']['bundle'] = [
          '#type' => 'container',
          '#prefix' => '<div id="' . $wrapper_id_bundle_fields . '">',
          '#suffix' => '</div>',
        ];
        if ($bundle) {
          $sub_form['fields']['bundle']['#type'] = 'checkboxes';
          $sub_form['fields']['bundle'] += [
            '#title' => $this->t('Bundle fields'),
            '#field_prefix' => $this->t('<div class="description fieldset__description">Please select the bundle fields to include into the denormalized table. Leave empty to include all of them.</div>'),
            '#default_value' => $configuration['fields']['bundle'] ?? [],
            '#options' => $this->getEntityBundleFieldsOptions($entity_type, $bundle),
          ];
        }
      }
    }
    else {
      $sub_form['base_table'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Base table'),
        '#default_value' => $configuration['base_table'] ?? '',
        '#required' => TRUE,
      ];
      $sub_form['changed_key'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Changed key'),
        '#default_value' => $configuration['changed_key'] ?? '',
      ];
      // @todo add ajax callback for db prefix
      $sub_form['external'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Is external?'),
        '#default_value' => $configuration['external'] ?? FALSE,
      ];
      // Add db prefix textfield when is eternal true.
    }

    return $sub_form;
  }

  /**
   * Get the bundle options for a given entity type.
   *
   * @param string $entity_type
   *   The entity type.
   *
   * @return array
   *   The bundle options.
   */
  private function getBundleOptions(string $entity_type): array {
    if (!$entity_type) {
      return [];
    }
    return array_map(function ($bundle_info) {
      return $bundle_info['label'];
    }, \Drupal::service('entity_type.bundle.info')->getBundleInfo($entity_type));
  }

  /**
   * Get the base fields options for a given entity type.
   *
   * @param string $entity_type_id
   *   The entity type ID.
   * @param bool $changed_key
   *   Whether to include the changed key.
   *
   * @return array
   *   The base fields options.
   */
  private function getEntityBaseFieldsOptions(string $entity_type_id, bool $changed_key = FALSE): array {
    if (!$entity_type_id) {
      return [];
    }
    $base_fields = $this->entityFieldManager()
      ->getBaseFieldDefinitions($entity_type_id);
    $options = [];
    foreach ($base_fields as $field) {
      $options[$field->getName()] = $field->getLabel();
    }
    return $options;
  }

  /**
   * Get the bundle fields options for a given entity type and bundle.
   *
   * @param string $entity_type_id
   *   The entity type ID.
   * @param string $bundle
   *   The bundle.
   *
   * @return array
   *   The bundle fields options.
   */
  private function getEntityBundleFieldsOptions(string $entity_type_id, string $bundle): array {
    if (!$entity_type_id || !$bundle) {
      return [];
    }

    $options = [];
    if ($entity_type_id === 'webform_submission' && ($webform = $this->getWebform($bundle))) {
      foreach ($webform->getElementsInitializedFlattenedAndHasValue() as $key => $element) {
        $title = isset($element['#title']) ? $this->t('@title', [
          '@title' => $element['#title'],
        ]) : $key;
        $options[$key] = $title;
      }
      return $options;
    }

    $entity_field_manager = $this->entityFieldManager();
    $field_definitions = $entity_field_manager->getFieldDefinitions($entity_type_id, $bundle);
    foreach ($field_definitions as $field) {
      if ($field->getTargetBundle()) {
        $options[$field->getName()] = $field->getLabel();
      }
    }
    return $options;
  }

  /**
   * Helper method to instantiate the entity field manager.
   *
   * @return \Drupal\Core\Entity\EntityFieldManagerInterface
   *   The entity field manager.
   */
  private function entityFieldManager(): EntityFieldManagerInterface {
    if (empty($this->entityFieldManager)) {
      $this->entityFieldManager = \Drupal::service('entity_field.manager');
    }
    return $this->entityFieldManager;
  }

  /**
   * Process the entity keys options.
   *
   * @param array $element
   *   The element.
   * @param string $entity_type_id
   *   The entity type ID.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function processEntityKeysOptions(array &$element, string $entity_type_id): void {
    $entity_keys = array_flip($this->entityTypeManager
      ->getDefinition($entity_type_id)
      ->getKeys());
    foreach ($element['#options'] as $key => $label) {
      if (isset($entity_keys[$key])) {
        $element[$key]['#disabled'] = TRUE;
        $element[$key]['#default_value'] = $key;
      }
    }
  }

  /**
   * Get the entity IDs to populate.
   *
   * @param \Drupal\denormalizer\Entity\DenormalizerTableInterface $dn_table
   *   The denormalizer table.
   * @param \Drupal\Core\Entity\EntityStorageInterface $entity_storage
   *   The entity storage.
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type_definition
   *   The entity type definition.
   *
   * @return array
   *   The entity IDs to populate.
   */
  private function getEntityIdsToPopulate(DenormalizerTableInterface $dn_table, EntityStorageInterface $entity_storage, EntityTypeInterface $entity_type_definition): array {
    $configuration = $dn_table->getSourceConfiguration();
    $bundle = $configuration['bundle'];
    $bundle_key = $entity_type_definition->getKey('bundle');

    $query = $entity_storage->getQuery()->accessCheck(FALSE);

    if ($bundle_key) {
      $query->condition($bundle_key, $bundle);
    }

    return $query->execute();
  }

  /**
   * Get the webform.
   *
   * @param string $webform_id
   *   The webform ID.
   *
   * @return \Drupal\webform\WebformInterface|null
   *   The webform.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function getWebform(string $webform_id): ?WebformInterface {
    /** @var \Drupal\webform\WebformEntityStorageInterface $webform_storage */
    $webform_storage = $this->entityTypeManager->getStorage('webform');
    /** @var \Drupal\webform\WebformInterface $webform */
    $webform = $webform_storage->load($webform_id);
    return $webform;
  }

}
